package MutualExclusion;

import daj.*;
import java.util.*;

public class RicartAgrawala extends Application {
    public static final int PNUM = 6;
    public static final int DISTANCE = 70;
    public static final int BORDER = 40;

    public static void main(String[] args) {
        new RicartAgrawala().run();
    }

    /** Creates new Main */
    public RicartAgrawala() {
        super("RicartAgrawala Mutual Exclusion",
            (int) ((PNUM * DISTANCE) / Math.PI) + 2 * BORDER, 
            (int) ((PNUM * DISTANCE) / Math.PI) + 2 * BORDER);
    }

    public void construct() {
        Node[] nodes = new Node[PNUM];
        int i, j;
        double phi, r = (PNUM * DISTANCE) / (2 * Math.PI);

        for (i=0; i<PNUM; i++) {
            phi = (double) i / PNUM * Math.PI * 2;
            nodes[i] = node(new Prog(i), new Integer(i).toString(),
                (int) (r * Math.cos(phi) + r + BORDER),
                (int) (r * Math.sin(phi) + r + BORDER));
        }

        for (i=0; i<PNUM; i++) {
            for (j=0; j<PNUM; j++) {
                link(nodes[i], nodes[j]);
                //link(nodes[j], nodes[i]);
            }
        }
    }

    public String getText() {
        return "An implementation of the RicartAgrawala Algorithm for\n" +
            "solving the distributed mutual exclusion problem.";
    }    
}

/**
 * The code that will be executed on each node.
 * @author  rene
 * @version
 */
class Prog extends Program {
    /** possible value for region: idle */
    public static final int R = 0;
    /** possible value for region: try to get the ressource */
    public static final int T = 1;
    /** possible value for region: critical region entered */
    public static final int C = 2;
    /** possible value for region: exiting critical region */
    //public static final int E = 3;
    
    /** the node number */
    protected int index;
    /** the region can be either R, T, C or E */
    protected int region = R;
    /** the current clock value */
    protected int clock = 0;
    /** for every node, the message history */
    protected Vector[] history = null;
    /** for every neighbour, the buffer for remembering deferred ok messages 
        until critical region has been left */
    protected Vector[] deferred_ok = null;
    /** the logical time of the last "try" message that has been sent */
    protected LogicalTime lastTryTime = null;
    
    /** the default constructor initializes the index and the arrays */
    public Prog(int index) {
        this.index = index;
        history = new Vector[RicartAgrawala.PNUM];
        for (int i=0; i<history.length; i++)
            history[i] = new Vector();
        deferred_ok = new Vector[RicartAgrawala.PNUM];
        for (int i=0; i<deferred_ok.length; i++)
            if (i != index)
                deferred_ok[i] = new Vector();
            else
                deferred_ok[i] = null;
    }
    
    /** the user function for trying to get the ressource (getting the 
        permission to enter the critical region) */
    public void tryRessource() {
        // single actions must be atomic --> synchronize them
        synchronized (this) {
            clock++;
            region = T;
            lastTryTime = new LogicalTime(clock, index);
            out().send(new Msg(Msg.TRY, lastTryTime));
        }
    }
    
    /** the user function for freeing the ressource (exiting the critical
        region) */
    public void freeRessource() {
        synchronized (this) {
            clock++;
            // this does not match the formal specification - the release is
            // done immediately
            region = R;
            for (int i=0; i<deferred_ok.length; i++)
                if (i != index) {
                    for (int j=0; j<deferred_ok[i].size(); j++) {
                        out(i).send((Msg) deferred_ok[i].get(j));
                    }
                    deferred_ok[i].removeAllElements();
                }
        }
    }
    
    /** this function is called when the ressource can be taken (the critical
        region can be entered) */
    public void ressourceAvailable() {
        // for testing purposes there is no user to ne notified
    }
    
    /** this function is called when the ressource has been freed (the critical
        region has been left) */
    public void ressourceFreed() {
        // for testing purposes there is no user to ne notified
    }

    /** this method performs the node's main functions */
    public void main() {
        Random rand = new Random();
        GlobalAssertion safety = new CriticalRegionConflict(),
                        liveness = new CriticalRegionLockout();

        while (true) {
            assert(safety);
            assert(liveness);
            /* first the tree construction routines */
            int received = in().select(2);
            if (received >= 0) {
                synchronized (this) {
                    Msg msg = (Msg) in(received).receive();
                    if (msg.time.clock > clock) 
                        clock = msg.time.clock;
                    clock++;
                    history[received].add(msg);
                    // respond to try requests from other nodes
                    if (msg.type == Msg.TRY && received != index) 
                        handleTryMessage(received, msg);
                    // check if the critical region can be entered now
                    if (region == T && checkOkMsgsReceived()) {
                        clock++;
                        region = C;
                        // the last try is invalid now
                        lastTryTime = null;
                    }
                }
            }
            // simulate the user here
            if (region == R && rand.nextInt(RicartAgrawala.PNUM * 2) == 0) {
                tryRessource();
            }
            else if (region == C && rand.nextInt(RicartAgrawala.PNUM) == 0) {
                freeRessource();
            }
        }
    }
    
    public String getText() {
        String s = "Node " + index + "\nRegion: ";
        switch (region) {
            case R: s += "idle";
                    break;
            case T: s += "trying to get lock:\n";
                    s += "try time = " + lastTryTime + "\n";
                    break;
            case C: s += "inside critical region";
                    break;
            /*case E: s += "exiting critical region";
                    break;*/
            default: s += "invalid";
        }
        s += "\nClock value: " + clock;
        return s;
    }
    
    /** This is a helper function that responds properly to a received "try"
        message. */
    private void handleTryMessage(int received, Msg msg) {
        Msg ackMsg = new Msg(Msg.OK, new LogicalTime(clock, index));
        if (region == R /*|| region == E*/) 
            out(received).send(ackMsg);
        else if (region == C) 
            deferred_ok[received].add(ackMsg);
        else if (region == T) {
            if (lastTryTime != null && msg.time.lessThan(lastTryTime)) 
                out(received).send(ackMsg);
            else 
                deferred_ok[received].add(ackMsg);
        }
    }
    
    /** This is a helper function that checks if all "OK" messages have been
        recieved for the last try message. */
    private boolean checkOkMsgsReceived() {
        boolean receivedOkMsgs = true;
        for (int i=0; i<history.length; i++) {
            if (i != index) {
                // always check the last "OK" msg of the history
                // queues, because this is the one that has been
                // recieved at last
                int j = history[i].size() - 1;
                while (j >= 0 && ((Msg) history[i].get(j)).type != Msg.OK)
                    j--;
                if (j >= 0) {
                    Msg lastMsg = (Msg) history[i].get(j);
                    if (! lastTryTime.lessThan(lastMsg.time)) 
                        // there is no "OK" message that is newer
                        // than our own "try" message
                        receivedOkMsgs = false;
                }
                else
                    // there is no "OK" message from this node
                    receivedOkMsgs = false;
            }
        }
        return receivedOkMsgs;
    }
}

/** Represents a logical time value with operations for checking for equality
  * and order.
  */
class LogicalTime {
    /** the clock value */
    public int clock;
    /** for equal clock values the sender process's index is compared */
    public int process;
    
    /** a logical time object can only be constructed when both parameters are
        known */
    public LogicalTime(int clock, int process) {
        this.clock = clock;
        this.process = process;
    }
    
    /** checks for equality as defined for the Lamport logical time */
    public boolean equals(Object o) {
        if (o instanceof LogicalTime && 
            ((LogicalTime) o).clock == clock &&
            ((LogicalTime) o).process == process)
            return true;
        else
            return false;
    }
    
    /** check if the the current time is less than t as defined for the Lamport
        logical time */
    public boolean lessThan(LogicalTime t) {
        if (clock < t.clock ||
            (clock == t.clock && process < t.process))
            return true;
        else
            return false;
    }
    
    public String toString() {
        return "(" + clock + ", " + process + ")";
    }
}

class Msg extends Message {
    /** message type: try (a process tries to get the shared ressource) */
    public static final int TRY = 0;
    /** message type: ok (a process confirms another process's request */
    public static final int OK = 1;
    
    /** the message type */
    public int type;
    /** the logical time of the sending process when the message was sent */
    public LogicalTime time;
    
    public Msg(int type, LogicalTime time) {
        this.type = type;
        this.time = time;
    }
    
    public boolean equals(Object o) {
        if (o instanceof Msg && 
            ((Msg) o).type == type &&
            ((Msg) o).time.equals(time))
            return true;
        else
            return false;
    }
    
    public String getText() {
        String s = "Message type: ";
        switch (type) {
            case TRY: s += "try";
                      break;
            case OK:  s += "ok";
                      break;
            default:  s += "invalid";
        }
        s += "\nTime: " + time.toString();
        return s;
    }
}

/** This assertion checks that at most 1 process is in the critical region at
    the same time. */
class CriticalRegionConflict extends GlobalAssertion {
    private Vector procsInCR = new Vector();
    
    public boolean assert(Program progs[]) {
        int numOfProcsInCR = 0;
        procsInCR.removeAllElements();
        
        for (int i=0; i<progs.length; i++) {
            if (((Prog) progs[i]).region == Prog.C) {
                numOfProcsInCR++;
                procsInCR.add(new Integer(i));
            }
        }
        return numOfProcsInCR <= 1;
    }
    
    public String getText() {
        String text = "Processes in the critical region: ";
        for (int i=0; i<procsInCR.size(); i++)
            text += procsInCR.get(i) + " ";
        return text;
    }
}

/** This assertion checks that if a process tries to enter the critical region,
    no other process trying it at any time later will be allowed to enter the
    critical region before the first process has entered it. */
class CriticalRegionLockout extends GlobalAssertion {
    private LogicalTime[] tryTimes = new LogicalTime[RicartAgrawala.PNUM];
    private int procInCR = -1, procTryingLonger = -1;
    
    public CriticalRegionLockout() {
        for (int i=0; i<tryTimes.length; i++)
            tryTimes[i] = null;
    }
    
    public boolean assert(Program progs[]) {
        // remember the times when the processes tried to enter the CR
        for (int i=0; i<progs.length; i++) 
            if (((Prog) progs[i]).region == Prog.T)
                tryTimes[i] = ((Prog) progs[i]).lastTryTime;
        
        // Now check when a process is in the CR, if another one is still 
        // trying but started to try earlier. This should not happen.
        for (int i=0; i<progs.length; i++)
            if (((Prog) progs[i]).region == Prog.C) {
                for (int j=0; j<progs.length; j++) {
                    if (((Prog) progs[j]).region == Prog.T &&
                        tryTimes[j].lessThan(tryTimes[i])) {
                            procInCR = i;
                            procTryingLonger = j;
                            return false;
                    }
                }
            }
            
        return true;
    }
    
    public String getText() {
        return "Process " + procInCR + " is in the critical region, but " +
            procTryingLonger + "was trying earlier !";
    }
}